Package: pedquant
Version: 0.2.6
Title: Public Economic Data and Quantitative Analysis
Description: 
    Provides an interface to access public economic and financial data for 
    economic research and quantitative analysis. The data sources including 
    NBS, FRED, Sina, Eastmoney and etc. It also provides quantitative 
    functions for trading strategies based on the 'data.table', 'TTR', 
    'PerformanceAnalytics' and etc packages.
Authors@R: 
    person("Shichen", "Xie", email = "xie@shichen.name", role = c("aut", "cre"))
Depends: R (>= 4.1.0)
Imports: data.table, TTR, zoo, PerformanceAnalytics, curl, httr, rvest,
        lubridate, stringi, jsonlite, readxl, readr, echarts4r, xefun
        (> 0.1.3), htmlwidgets
Suggests: knitr, rmarkdown
License: GPL-3
URL: https://github.com/ShichenXie/pedquant, https://pedquant.com/
BugReports: https://github.com/ShichenXie/pedquant/issues
LazyData: true
RoxygenNote: 7.3.2
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2025-09-24 13:26:28 UTC; shichenxie
Author: Shichen Xie [aut, cre]
Maintainer: Shichen Xie <xie@shichen.name>
Repository: CRAN
Date/Publication: 2025-09-24 14:00:02 UTC
Built: R 4.6.0; ; 2026-04-30 15:00:19 UTC; unix
