# Generated by roxygen2: do not edit by hand

export("%x<%")
export("%x>%")
export(ed_code)
export(ed_fred)
export(ed_fred_symbol)
export(ed_nbs)
export(ed_nbs_subregion)
export(ed_nbs_symbol)
export(md_bond)
export(md_forex)
export(md_future)
export(md_future_symbol)
export(md_money)
export(md_moneycn)
export(md_stock)
export(md_stock_adjust)
export(md_stock_financials)
export(md_stock_symbol)
export(md_symbol)
export(pq_addti)
export(pq_addti_funs)
export(pq_freq)
export(pq_opr)
export(pq_performance)
export(pq_performance_funs)
export(pq_plot)
export(pq_portfolio)
export(pq_return)
import(PerformanceAnalytics)
import(TTR)
import(data.table)
import(echarts4r)
import(httr)
import(xefun)
importFrom(curl,curl)
importFrom(curl,curl_download)
importFrom(curl,handle_setheaders)
importFrom(curl,has_internet)
importFrom(curl,new_handle)
importFrom(curl,nslookup)
importFrom(htmlwidgets,JS)
importFrom(httr,POST)
importFrom(httr,add_headers)
importFrom(httr,set_config)
importFrom(httr,use_proxy)
importFrom(httr,write_disk)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,read_json)
importFrom(lubridate,as_date)
importFrom(readr,read_csv)
importFrom(readr,read_lines)
importFrom(readxl,read_excel)
importFrom(rvest,"%>%")
importFrom(rvest,html_attr)
importFrom(rvest,html_nodes)
importFrom(rvest,html_table)
importFrom(rvest,html_text)
importFrom(rvest,read_html)
importFrom(stats,IQR)
importFrom(stats,lm)
importFrom(stats,na.omit)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,weighted.mean)
importFrom(stringi,stri_unescape_unicode)
importFrom(utils,data)
importFrom(utils,download.file)
importFrom(utils,menu)
importFrom(utils,read.csv)
importFrom(xefun,as.list2)
importFrom(xefun,date_bop)
importFrom(xefun,date_eop)
importFrom(xefun,date_num)
importFrom(zoo,na.locf0)
